#include "pre.h"

// For the macromodel
#include "mcgate.h"
#include "mtellnp.h"
#include "mnode.h"
#include "mbranch.h"
#include "mgroup.h"
// For the circuit
#include "cgate.h"
#include "tellnp.h"
#include "node.h"
#include "branch.h"
#include "group.h"
#include "macromodel.h"
#include "patternmatch.h"

#include "wire.h"

#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <string>
#include <cstring>
#include <iostream>
#include <sys/times.h>

#define NLAlloc(type,num) (type *)calloc(num,sizeof(type))
#define ionumber 500

extern int yyerror(char const *);
extern int yywrap();
extern int yyparse();
extern FILE *yyin;

Element *ElementHead;
Action *ActionHead;
ModelList *ModelHead;
BipolarList *BipolarHead;
MacroModelList *MacroModelHead;
ResultList *ResultHead;

MosList *MosHead;
NodeList *NodeHead;
TransList *TransHead;
WireList *WireHead;

InoutList *InputHead, *OutputHead;

int device_biggest;
int node_biggest;
int trans_biggest;

PMosList *hash_mos_list;
PNodeList *hash_node_list;

int num_of_gates;
int num_of_trans;
PCGate *hash_gate_list;
PTransList *hash_trans_list;

int file_num;
char **files;
extern int flag;

double UnixTime();
bool macromodel_head_done();
MacroModelList *return_macromodel();


int main(int argc, char *argv[]) 
{
    double time, time_begin, time_end, gate_finish, patt_begin;
	char *pvdd = NULL;
    char *pgnd = NULL;


    time_begin = UnixTime();

	struct tms buffer;
	double totaluser, totalsys, prevuser, prevsys, miduser, midsys;
	double parseruser, parsersys, tellnpuser, tellnpsys, nodeuser, nodesys, branchuser, branchsys, gbranchuser, gbranchsys, prerecoguser, prerecogsys, nprecoguser, nprecogsys, coupleuser, couplesys, gaterecoguser, gaterecogsys, preconnectuser, preconnectsys, gateconnectuser, gateconnectsys;

/** Measure time using SubGemini method ***/
    times(&buffer);
    prevuser = buffer.tms_utime;
    prevsys = buffer.tms_stime;
//    printf("Setup Time:\n\tUser %d ", buffer.tms_utime / 6 );
//    printf("\n\tSystem %d\n", buffer.tms_stime / 6);
/*****************************************/
    file_num=1;
    files = argv;

    FILE *fp_out;
    FILE *fp_header;

    if(argc>4)
    {
	yyin = fopen(argv[1], "r");
	if(yyin == NULL)
	{
	    perror("can not open source file !\n");
	    return -1;
	}
	fp_out = fopen(argv[3], "w");
	if(fp_out==NULL) 
	{
	    perror("can not open destination file !\n");
	    return -1;
	}
	fp_header = fopen(argv[4], "w");
	if(fp_header==NULL) 
	{
	    perror("can not open destination file !\n");
	    return -1;
	}
    }
    else
    {
	printf("Useage: parser input output \n");
	return 0;
    }

    ElementHead = NULL;
    ActionHead = NULL;
    ModelHead = NULL;
    MosHead = NULL;
	NodeHead = NULL;
    InputHead = NULL;
    OutputHead = NULL;
    TransHead = NULL;
	WireHead = NULL;
	BipolarHead = NULL;
	ResultHead = NULL;
	MacroModelHead = NULL;

    flag = 1;

    yyparse();

    times(&buffer);
    parseruser = buffer.tms_utime;
    parsersys = buffer.tms_stime;

/****** Add p or n character of Mosfet netlist *******/
    TellNP *tellnp;
    tellnp = new TellNP();
    tellnp->mos_np_netlist();
	node_biggest = tellnp->return_node_biggest();
	device_biggest = tellnp->return_device_biggest();

	printf("node_beggest: %d,  device_biggest: %d\n", node_biggest, device_biggest);
    
/*
    for(MosList *cur_m=MosHead; cur_m!=NULL; cur_m=cur_m->next)
	printf("name=%s d=%s, g=%s, s=%s\n",cur_m->mosname,cur_m->d, cur_m->g, cur_m->s);
    printf("\n");
*/

        hash_mos_list = new PMosList[device_biggest];
        hash_node_list = new PNodeList[node_biggest+ionumber];
        for(int i=0; i<device_biggest; i++)
                hash_mos_list[i] = NULL;
        for(int j=0; j<(node_biggest+ionumber); j++)
                hash_node_list[j] = NULL;

	printf("TellNP part done\n");

    times(&buffer);
    tellnpuser = buffer.tms_utime;
    tellnpsys = buffer.tms_stime;


/****** Genarate NodeList ***************************/
    Node *mynode;
	mynode = new Node();

	printf("hehe_1\n");

    mynode->gen_nodelist();
	printf("hehe_2\n");

    mynode->gen_mos_list();
	printf("hehe_3\n");


/*
	for(int i=0; i<(device_biggest+1); i++)
	{
	  if(hash_mos_list[i]!=NULL)
	  {
		printf(" %s(d=%s,g=%s,s=%s)\n", hash_mos_list[i]->mosname, hash_mos_list[i]->d, hash_mos_list[i]->g, hash_mos_list[i]->s);
	  }
	  else
		printf(" %d -- NULL \n", i);
	}
*/

    mynode->gen_np_char_netlist();
	printf("hehe_3_5\n");

/*
	for(int i=0; i<(node_biggest+ionumber); i++)
	{
	  if(hash_node_list[i]!=NULL)
	  {
		printf(" %d -------- %s\n", i, hash_node_list[i]->nodename);
		for(DeviceList *curr=hash_node_list[i]->devicelist; curr!=NULL; curr=curr->next)
		{
			printf(" %s(%d) \n", curr->devicename,curr->pintype);
		}
		printf(" pn_node ? = %d \n", hash_node_list[i]->pn_node);
	  }
	  else
		printf(" %d -- NULL \n", i);
	}
*/

/*
	cout<<" Input head: "<<endl;
	for(InoutList *curr=InputHead; curr!=NULL; curr=curr->next)
	{
		cout<<" In: "<<curr->pinname;
	}
	cout<<endl;
*/

    mynode->delete_vdd_gnd_from_inputlist();
	pvdd = mynode->get_vdd(MosHead);
	pgnd = mynode->get_gnd(MosHead);
	printf("hehe_4\n");


    mynode->generate_output_pinlist();
//	printf("hehe_5\n");

    mynode->generate_input_pinlist();
//	printf("hehe_6\n");
 /* 
    for(NodeList *cur_n=NodeHead; cur_n!=NULL; cur_n=cur_n->next)
    {
	printf("%s(%d): ", cur_n->nodename,cur_n->pn_node);
	for(DeviceList *cur_d=cur_n->devicelist; cur_d!=NULL; cur_d=cur_d->next)
	    printf("%s ",cur_d->devicename);
	printf("\n");
    }
*/
	printf("Node part done\n");

    times(&buffer);
    nodeuser = buffer.tms_utime;
    nodesys = buffer.tms_stime;
/****** Generate 4 moslist *************************/
    Group *mygroup;
    mygroup = new Group();
    mygroup->generate_branch_list();
	printf("Generate_branch_list\n");

    times(&buffer);
    branchuser = buffer.tms_utime;
    branchsys = buffer.tms_stime;




/*
    printf("N_Branch_list\n");
    Branch *cur_b=mygroup->n_branch_list;
    if(cur_b==NULL)
	printf("empty\n");
    else if(cur_b->next==NULL)
	printf("only one\n");
    else {
    for(; cur_b!=NULL; cur_b=cur_b->next)
    {
	printf("n_branch_list_i:\n");
	printf("n_or_p=%d, success=%d, pn_node=%s\n",cur_b->branch_type,cur_b->success,cur_b->pn_node);
	printf("Moslist: ");
	for(MosList *cur_mo=cur_b->mos_list; cur_mo!=NULL; cur_mo=cur_mo->next)
	    printf("%s  ", cur_mo->mosname);
	printf("\n");
	printf("Nodelist: ");
	if(cur_b->node_list==NULL)
	    printf("No internal node\n");
	else
	{
	for(BranchNodeList *cur_no=cur_b->node_list; cur_no!=NULL; cur_no=cur_no->next)
	    printf("%s  ",cur_no->nodename);
	printf("\n");
	}
    }
    }

    printf("P_Branch_list\n");
    Branch *cur_bb=mygroup->p_branch_list;
    if(cur_bb==NULL)
        printf("empty\n");
    else if(cur_bb->next==NULL)
        printf("only one\n");
    else {
    for(; cur_bb!=NULL; cur_bb=cur_bb->next)
    {
        printf("n_branch_list_i:\n");
        printf("n_or_p=%d, success=%d, pn_node=%s\n",cur_bb->branch_type,cur_bb->success,cur_bb->pn_node);
        printf("Moslist: ");
        for(MosList *cur_mo=cur_bb->mos_list; cur_mo!=NULL; cur_mo=cur_mo->next)
            printf("%s  ", cur_mo->mosname);
        printf("\n");
        printf("Nodelist: ");
        if(cur_bb->node_list==NULL)
            printf("No internal node\n");
        else
        {
        for(BranchNodeList *cur_no=cur_bb->node_list; cur_no!=NULL; cur_no=cur_no->next)
            printf("%s  ",cur_no->nodename);
        printf("\n");
	}
    }
    }
*/


/*
	for(int i=0; i<device_biggest; i++)
	{
	  if(hash_mos_list[i]!=NULL)
	  {
		printf(" %s(d=%s,g=%s,s=%s) in_branch=%d\n", hash_mos_list[i]->mosname, hash_mos_list[i]->d, hash_mos_list[i]->g, hash_mos_list[i]->s, hash_mos_list[i]->in_branch);
	  }
	  else
		printf(" %d -- NULL \n", i);
	}
*/



    mygroup->group_branch_list();
	printf("Group_branch_list\n");

    times(&buffer);
    gbranchuser = buffer.tms_utime;
    gbranchsys = buffer.tms_stime;


/*
    printf("N_Branch_list\n");
    Branch *cur_b=mygroup->n_branch_list;
    if(cur_b==NULL)
        printf("empty\n");
    else {
    for(; cur_b!=NULL; cur_b=cur_b->next)
    {
        printf("n_branch_list_i:\n");
        printf("n_or_p=%d, success=%d, pn_node=%s\n",cur_b->branch_type,cur_b->success,cur_b->pn_node);
        printf("Moslist: ");
        for(MosList *cur_mo=cur_b->mos_list; cur_mo!=NULL; cur_mo=cur_mo->next)
            printf("%s  ", cur_mo->mosname);
        printf("\n");
        printf("Nodelist: ");
        if(cur_b->node_list==NULL)
            printf("No internal node\n");
        else
        {
        for(BranchNodeList *cur_no=cur_b->node_list; cur_no!=NULL; cur_no=cur_no->next)
            printf("%s  ",cur_no->nodename);
        printf("\n");
        }

    }
    }

    printf("P_Branch_list\n");
    Branch *cur_bb=mygroup->p_branch_list;
    if(cur_bb==NULL)
        printf("empty\n");
    else if(cur_bb->next==NULL)
        printf("only one\n");
    else {
    for(; cur_bb!=NULL; cur_bb=cur_bb->next)
    {
        printf("n_branch_list_i:\n");
        printf("n_or_p=%d, success=%d, pn_node=%s\n",cur_bb->branch_type,cur_bb->success,cur_bb->pn_node);
        printf("Moslist: ");
        for(MosList *cur_mo=cur_bb->mos_list; cur_mo!=NULL; cur_mo=cur_mo->next)
            printf("%s  ", cur_mo->mosname);
        printf("\n");
        printf("Nodelist: ");
        if(cur_bb->node_list==NULL)
            printf("No internal node\n");
        else
        {
        for(BranchNodeList *cur_no=cur_bb->node_list; cur_no!=NULL; cur_no=cur_no->next)
            printf("%s  ",cur_no->nodename);
        printf("\n");
        }

    }
    }
*/




    mygroup->generate_pass_left_list();
	printf("generate_pass_left_list\n");

/*
    for(MosList *curr_mo=mygroup->pass_transistor_list; curr_mo!=NULL; curr_mo=curr_mo->next)
	printf("%s->", curr_mo->mosname);
    printf("\n");
*/

    mygroup->generate_trans_list();
	printf("generate_trans_list\n");
    
/*
    for(TransList *curr_t=TransHead; curr_t!=NULL; curr_t=curr_t->next)
    {
	printf("begin:\n");
	printf("input=%s, output=%s, p_g=%s, n_g=%s, delete=%d\n",  curr_t->input, curr_t->output, curr_t->p_g, curr_t->n_g, curr_t->ready_to_delete);
	printf("end.\n");
    }
*/

// Added function here.
	mygroup->generate_num_for_branch();
	printf("Generate num for branch\n");

    times(&buffer);
    prerecoguser = buffer.tms_utime;
    prerecogsys = buffer.tms_stime;


    mygroup->n_branch_recognition();
	printf("n branch recognition\n");
    mygroup->p_branch_recognition();
	printf("p branch recognition\n");

    times(&buffer);
    nprecoguser = buffer.tms_utime;
    nprecogsys = buffer.tms_stime;

/*
    for(Branch *cc=mygroup->n_branch_list; cc!=NULL; cc=cc->next)
    {
	printf("(%s)->%s\n", cc->pn_node,cc->headername);
    }

    printf("N_Branch_list(recognized)\n");
    for(Branch *cur_b=mygroup->n_branch_list; cur_b!=NULL; cur_b=cur_b->next)
    {
	MosList *mosm = cur_b->mos_list;
	printf("Branch (%s)'s function is %s \n", cur_b->pn_node, mosm->g);
    }
    printf("P_Branch_list(recognized)\n");
    for(Branch *cur_b=mygroup->p_branch_list; cur_b!=NULL; cur_b=cur_b->next)
    {
	MosList *mosm = cur_b->mos_list;
	printf("Branch (%s)'s function is %s \n", cur_b->pn_node, mosm->g);
    }
*/
	mygroup->give_name_to_branch();
		cout<<"give name\n";
	mygroup->gen_hash_branch_list();
		cout<<"gen_hash_branch\n";

	mygroup->couple_p_n_branch();
		cout<<"couple\n";

    times(&buffer);
    coupleuser = buffer.tms_utime;
    couplesys = buffer.tms_stime;


/*
	for(int i=0; i<(node_biggest+ionumber); i++)
	{
	  if(hash_node_list[i]!=NULL)
	  {
		printf(" %d --------------------------- %s\n", i, hash_node_list[i]->nodename);
		for(DeviceList *curr=hash_node_list[i]->devicelist; curr!=NULL; curr=curr->next)
		{
			printf(" %s(%d) \n", curr->devicename,curr->pintype);
		}
		printf(" pn_node ? = %d \n", hash_node_list[i]->pn_node);

		if(hash_node_list[i]->connected_p_branch_name != NULL)
			cout<<"p_branch name: "<<hash_node_list[i]->connected_p_branch_name<<endl;
		else
			cout<<"p_branch name: NULL"<<endl;

		if(hash_node_list[i]->connected_n_branch_name != NULL)
			cout<<"n_branch name: "<<hash_node_list[i]->connected_n_branch_name<<endl;
		else
			cout<<"n_branch name: NULL"<<endl;
	  }
	  else
		printf(" %d -- NULL \n", i);
	}
*/
	
	mygroup->gate_recognition();
	printf("Gate recognition\n");

    times(&buffer);
    gaterecoguser = buffer.tms_utime;
    gaterecogsys = buffer.tms_stime;

/*
  if(mygroup->cmos_gate_list==NULL)
	 cout<<"NULL\n";
  else
  {
		cout<<"here\n";
	for(CGate *cup = mygroup->cmos_gate_list; cup!=NULL; cup=cup->next)
	{
		if(cup->gatename != NULL)
			cout<<" A new gate ....................."<<cup->gatename<<endl;
		else
			cout<<" A new gate ..................... name = NULL\n";

		cout<<"pn_node: "<<(cup->n_branch)->pn_node;
		cout<<"real_gate: "<<cup->cmos_gate<<"p_fun: "<<(cup->p_branch)->headername<<" n_fun: "<<(cup->n_branch)->headername<<endl;
	}
  }
*/

    mygroup->gen_hash_gate_list();
		cout<<" num of gates "<<num_of_gates<<endl;
	printf("gen hash gate list\n");

		cout<<num_of_gates<<" duoshaoge "<<endl;
/*
	for(int i=0; i<num_of_gates; i++)
	{
		if(hash_gate_list[i]!=NULL)
		{
//		  if(!strcmp(hash_gate_list[i]->gatename, "gate655") || !strcmp(hash_gate_list[i]->gatename, "gate1102")  )
//		  {
			cout<<" A new gate ....................."<<hash_gate_list[i]->gatename<<endl;
			cout<<"pn_node: "<<(hash_gate_list[i]->n_branch)->pn_node<<endl;
			cout<<"real_gate: "<<hash_gate_list[i]->cmos_gate<<"  p_fun: "<<(hash_gate_list[i]->p_branch)->headername<<" n_fun: "<<(hash_gate_list[i]->n_branch)->headername<<endl;
			cout<<" Relation list: : :\n";
			if((hash_gate_list[i]->n_branch)->relation_list == NULL)
				cout<<"relation list  NULL\n";
			for(RelationList *cr=(hash_gate_list[i]->n_branch)->relation_list; cr!=NULL; cr=cr->next)
			{
				cout<<"type: "<<cr->type;
				for(GateNameList *cn=cr->gatename_list; cn!=NULL; cn=cn->next)
					cout<<" name:  "<<cn->name;
				cout<<endl;
			}
//		  }
		}
		else
			cout<<" Gate here ..................................NULL\n";
	}
*/


	mygroup->gen_inlist_for_gate();
	printf("gen inlist for every gate\n");

/*
	for(int i=0; i<num_of_gates; i++)
	{
		if(hash_gate_list[i]!=NULL)
		{
		    for(PinList *cu=hash_gate_list[i]->inlist; cu!=NULL; cu=cu->next)
		    {
			    cout<<" "<<cu->NodeName;
		    }
		    	cout<<endl;
		}
	}
*/



	mygroup->gen_hash_trans_list();
	printf("gen hash trans list\n");
	cout<<" Num of Gates: "<<num_of_gates<<" Num of trans: "<<num_of_trans<<endl;

/*
	for(int i=0; i<(num_of_trans+1); i++)
	{
		if(hash_trans_list[i]!=NULL)
		{
			cout<<" A new gate ....................."<<hash_trans_list[i]->transname<<endl;
			cout<<"input: "<<hash_trans_list[i]->input<<"  output: "<<hash_trans_list[i]->output<<endl;

		}
	}
*/
    times(&buffer);
    preconnectuser = buffer.tms_utime;
    preconnectsys = buffer.tms_stime;


	mygroup->gen_connection_of_gates();
	mygroup->gen_connection_of_trans();
	printf("gen connection\n");

    times(&buffer);
    gateconnectuser = buffer.tms_utime;
    gateconnectsys = buffer.tms_stime;


/*
	for(int i=0; i<(node_biggest+ionumber); i++)
	{
		if(hash_node_list[i]!=NULL)
		{
			if(!strcmp(hash_node_list[i]->nodename,"1649") || !strcmp(hash_node_list[i]->nodename,"2510"))
			{
			cout<<endl;
			cout<<"node ............................."<<hash_node_list[i]->nodename<<endl;
			if(hash_node_list[i]->output_of_gate_name != NULL)
				cout<<"  Output of gate: "<<hash_node_list[i]->output_of_gate_name;
			else
				cout<<"  No output gate";

			if(hash_node_list[i]->input_of_gate_list != NULL)
			{
				cout<<"  input of gate: ";
				for(PinList *curr=hash_node_list[i]->input_of_gate_list; curr!=NULL; curr=curr->next)
					cout<<" "<<curr->NodeName;
				cout<<endl;
			}
			else
				cout<<"  No inputput gate";

			
			if(hash_node_list[i]->input_of_trans_name!=NULL)
				cout<<" Input of trans: "<<hash_node_list[i]->input_of_trans_name<<endl;
			else
				cout<<" Input of trans: NULL"<<endl;
			if(hash_node_list[i]->output_of_trans_name!=NULL)
				cout<<" Output of trans: "<<hash_node_list[i]->output_of_trans_name<<endl;
			else
				cout<<" Output of trans: NULL"<<endl;
		}
		}
		else
		{
//			cout<<" Node ........................ NULL\n";
			;
		}
	}
*/


	mygroup->count_input_num_of_node();
	printf("Count the input num of every node \n");

/*
	for(int i=0; i<=(node_biggest+ionumber); i++)
	{
		if(hash_node_list[i]!=NULL)
		{
			cout<<"node name: "<<hash_node_list[i]->nodename<<"  num: "<<hash_node_list[i]->num_of_inputs<<endl;
		}
	}
*/
	mygroup->gen_num_for_macromodel();
	cout<<"Gen num for every macromodel\n";


	cout<<" Circuit part done. \n";


	cout<<" output: macromodel \n";
    for(MacroModelList *curr_model=MacroModelHead; curr_model!=NULL; curr_model=curr_model->next)
	{
		cout<<" name: "<<curr_model->macromodel_name<<" num of mos: "<<curr_model->num_of_mos<<"  done: "<<curr_model->done<<endl;
	}

    gate_finish = UnixTime();

// Measure the phase 1 time
      times(&buffer);
      miduser = buffer.tms_utime;
      midsys = buffer.tms_stime;

/***************** Patternmatch work ********************************/
	printf("************** Begin the macromodel **************\n");

    patt_begin = UnixTime();

  while(!macromodel_head_done())
  {
/*
    for(MacroModelList *curr_model=MacroModelHead; curr_model!=NULL; curr_model=curr_model->next)
	{
	  if(!curr_model->done)
	  {
		if(curr_model->num_of_mos > here);
		{
			here = curr_model->num_of_mos;
			hmac = curr_model;
		}
	  }
	}
*/


		MacroModelList *hmac = return_macromodel();
		cout<<" begin pattern matching for ............ : "<<hmac->macromodel_name<<endl;
		Macromodel *a_macromodel = new Macromodel(hmac);
		a_macromodel->gen_data_structure();

/*
if(!strcmp(hmac->macromodel_name, "DFFP"))
{
	for(int i=0; i<(num_of_gates+1); i++)
	{
		if((a_macromodel->hash_gate_list[i])!=NULL)
		{
			cout<<" A new gate ....................."<<(a_macromodel->hash_gate_list[i])->gatename<<endl;
			cout<<"pn_node: "<<((a_macromodel->hash_gate_list[i])->n_branch)->pn_node<<endl;
			cout<<"real_gate: "<<(a_macromodel->hash_gate_list[i])->cmos_gate<<"  p_fun: "<<((a_macromodel->hash_gate_list[i])->p_branch)->headername<<" n_fun: "<<((a_macromodel->hash_gate_list[i])->n_branch)->headername<<endl;
			cout<<" N_branch gate: ";
			for(MosList *ci=((a_macromodel->hash_gate_list[i])->n_branch)->mos_list; ci!=NULL; ci=ci->next)
				cout<<" Mos: "<<ci->mosname;
			cout<<endl;
			cout<<" P_branch gate: ";
			for(MosList *ci=((a_macromodel->hash_gate_list[i])->p_branch)->mos_list; ci!=NULL; ci=ci->next)
				cout<<" Mos: "<<ci->mosname;
			cout<<endl;

		}
	}
}
*/
		
		cout<<" begin patternmatching: \n";
		Patternmatch *mypatternmatch = new Patternmatch(a_macromodel);
		mypatternmatch->pattern_match_process();
	}

/********** Output the gatelist after patternmatch *****************/
/*
	for(int i=0; i<(node_biggest+ionumber); i++)
	{
		if(hash_node_list[i]!=NULL)
		{
			cout<<endl;
			cout<<"node ............................."<<hash_node_list[i]->nodename<<endl;
			if(hash_node_list[i]->output_of_gate_name != NULL)
				cout<<"  Output of gate: "<<hash_node_list[i]->output_of_gate_name;
			else
				cout<<"  output gate:  NULL";

			if(hash_node_list[i]->input_of_gate_list != NULL)
			{
				cout<<"  input of gate: ";
				for(PinList *curr=hash_node_list[i]->input_of_gate_list; curr!=NULL; curr=curr->next)
					cout<<" "<<curr->NodeName;
				cout<<endl;
			}
			else
				cout<<"  inputput gate: NULL";

			
			if(hash_node_list[i]->input_of_trans_name!=NULL)
				cout<<" Input of trans: "<<hash_node_list[i]->input_of_trans_name<<endl;
			else
				cout<<" Input of trans: NULL"<<endl;
			if(hash_node_list[i]->output_of_trans_name!=NULL)
				cout<<" Output of trans: "<<hash_node_list[i]->output_of_trans_name<<endl;
			else
				cout<<" Output of trans: NULL"<<endl;
		}
		else
			cout<<" Node here is ..................... NULL\n";
	}
*/
/**************** Generate the WireHead ******************************/
	
	Wire *gwire = new Wire();
	gwire->generate_wirelist();
	cout<<" Wire gen ended\n";

/***** Output "header.v"  **********************************************/

    fprintf(fp_header, " /*****************************************************************/\n ");
    fprintf(fp_header, "/***          Static CMOS Circuit Extractor Program            ***/\n ");
    fprintf(fp_header, "/***          Output Header.v for Verilog Simulation           ***/\n");
    fprintf(fp_header, " /*****************************************************************/\n ");
    fprintf(fp_header, "\n");


    fprintf(fp_header, "module TRANS(in, out);\n");
    fprintf(fp_header, "input in;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = in;\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module INV(a, out);\n");
    fprintf(fp_header, "input a;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~a;\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module NAND2(a, b, out);\n");
    fprintf(fp_header, "input a, b;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~(a & b);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module NOR2(a, b, out);\n");
    fprintf(fp_header, "input a, b;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~(a | b);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module NAND3(a, b, c, out);\n");
    fprintf(fp_header, "input a, b, c;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~(a & b & c);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module NOR3(a, b, c, out);\n");
    fprintf(fp_header, "input a, b, c;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~(a | b | c);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module AOI12(a, b, c, out);\n");
    fprintf(fp_header, "input a, b, c;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a & b) | c);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module OAI12(a, b, c, out);\n");
    fprintf(fp_header, "input a, b, c;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a | b) & c);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module NAND4(a, b, c, d, out);\n");
    fprintf(fp_header, "input a, b, c, d;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~(a & b & c & d);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module NOR4(a, b, c, d, out);\n");
    fprintf(fp_header, "input a, b, c, d;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~(a | b | c | d);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module AOI22(a, b, c, d, out);\n");
    fprintf(fp_header, "input a, b, c, d;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a & b) | (c & d));\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module OAI22(a, b, c, d, out);\n");
    fprintf(fp_header, "input a, b, c, d;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a | b) & (c | d));\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module AOI13(a, b, c, d, out);\n");
    fprintf(fp_header, "input a, b, c, d;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a & b & c) | d);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module OAI13(a, b, c, d, out);\n");
    fprintf(fp_header, "input a, b, c, d;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a | b | c) & d);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module AOOI112(a, b, c, d, out);\n");
    fprintf(fp_header, "input a, b, c, d;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a & b) | c | d);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module OAAI112(a, b, c, d, out);\n");
    fprintf(fp_header, "input a, b, c, d;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a | b) & c & d);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module AOAI112(a, b, c, d, out);\n");
    fprintf(fp_header, "input a, b, c, d;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~(((a & b) | c) & d);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module OAOI112(a, b, c, d, out);\n");
    fprintf(fp_header, "input a, b, c, d;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~(((a | b) & c) | d);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module AOI23(a, b, c, d, e, out);\n");
    fprintf(fp_header, "input a, b, c, d, e;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a & b & c) | (d & e));\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module AOI311(a, b, c, d, e, out);\n");
    fprintf(fp_header, "input a, b, c, d, e;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a | b) | (c & d & e));\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module AOI221(a, b, c, d, e, out);\n");
    fprintf(fp_header, "input a, b, c, d, e;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a & b) | (c & d) | e);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module OAI23(a, b, c, d, e, out);\n");
    fprintf(fp_header, "input a, b, c, d, e;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a | b | c) & (d | e));\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module OAI311(a, b, c, d, e, out);\n");
    fprintf(fp_header, "input a, b, c, d, e;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a | b | c) & d & e);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module OAI221(a, b, c, d, e, out);\n");
    fprintf(fp_header, "input a, b, c, d, e;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a | b) & (c | d) & e);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module OAOI212(a, b, c, d, e, out);\n");
    fprintf(fp_header, "input a, b, c, d, e;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~(((a | b) & c) | (d & e));\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module OAOI2111(a, b, c, d, e, out);\n");
    fprintf(fp_header, "input a, b, c, d, e;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~(((a | b) & e) | (c | d));\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module AOAI212(a, b, c, d, e, out);\n");
    fprintf(fp_header, "input a, b, c, d, e;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a | b) & ((c & d) | e));\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module AOAI2111(a, b, c, d, e, out);\n");
    fprintf(fp_header, "input a, b, c, d, e;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~(((a & b) | e) & (c & d));\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module AOI33(a, b, c, d, e, f, out);\n");
    fprintf(fp_header, "input a, b, c, d, e, f;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a & b & c) | (d & e & f));\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module AOI321(a, b, c, d, e, f, out);\n");
    fprintf(fp_header, "input a, b, c, d, e, f;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a & b & c) | (d & e) | f);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module AOI222(a, b, c, d, e, f, out);\n");
    fprintf(fp_header, "input a, b, c, d, e, f;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a & b) | ( c & d) | (e & f));\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module OAI222(a, b, c, d, e, f, out);\n");
    fprintf(fp_header, "input a, b, c, d, e, f;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a | b) & (c | d) & (e | f));\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module OAI33(a, b, c, d, e, f, out);\n");
    fprintf(fp_header, "input a, b, c, d, e, f;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a | b | c) & (d & e & f));\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");

    fprintf(fp_header, "module OAI321(a, b, c, d, e, f, out);\n");
    fprintf(fp_header, "input a, b, c, d, e, f;\n");
    fprintf(fp_header, "output out;\n");
    fprintf(fp_header, "assign out = ~((a | b | c) & (d | e) & f);\n");
    fprintf(fp_header, "endmodule\n");
    fprintf(fp_header, "\n");


/*************** End of output of the header.v ******************************/

/***** Output Result ******************************/
    fprintf(fp_out, " /*****************************************************************/\n ");
    fprintf(fp_out, "/***          Static CMOS Circuit Extractor Program            ***/\n ");
    fprintf(fp_out, "/***        Output verilog netlist from Hspice netlist         ***/\n");
    fprintf(fp_out, " /*****************************************************************/\n ");
    fprintf(fp_out, "\n");

    InoutList *circuitname = InputHead;
    fprintf(fp_out, " module %s (", circuitname->pinname);
    for(InoutList *cur=circuitname->next; cur!=NULL; cur=cur->next)
	fprintf(fp_out, "%s, ", cur->pinname);

	fprintf(fp_out, "%s, %s, ", pvdd, pgnd); // Add VDD, GND in port descrip.

    for(InoutList *cur_out=OutputHead; cur_out!=NULL; cur_out=cur_out->next)
    {
	if(cur_out->next==NULL)
	    fprintf(fp_out, "%s", cur_out->pinname);
	else
	    fprintf(fp_out, "%s, ", cur_out->pinname);
    }
    fprintf(fp_out, "); \n");

    fprintf(fp_out, " input ");

	fprintf(fp_out, "%s, %s, ", pvdd, pgnd); // Add VDD, GND in input_list.

    for(InoutList *cur_1=InputHead->next; cur_1!=NULL; cur_1=cur_1->next)
    {   
	if(cur_1->next==NULL)
	    fprintf(fp_out, "%s", cur_1->pinname);
	else
	    fprintf(fp_out, "%s, ", cur_1->pinname);
    }
    fprintf(fp_out, ";\n");

	cout<<" Output input \n";

    fprintf(fp_out, " output ");
    for(InoutList *cur_2=OutputHead; cur_2!=NULL; cur_2=cur_2->next)
    {
	if(cur_2->next==NULL)
	    fprintf(fp_out, "%s", cur_2->pinname);
	else
	    fprintf(fp_out, "%s, ", cur_2->pinname);
    }
    fprintf(fp_out, ";\n");

	cout<<" Output output \n";

    fprintf(fp_out," wire ");
	int new_line=0;
    for(WireList *curr_w=WireHead; curr_w!=NULL; curr_w=curr_w->next)
    {
	  new_line+=1;
		if(curr_w->next==NULL)
		{
	    		fprintf(fp_out, " %s",curr_w->wirename);
		}
		else
		{
	    		fprintf(fp_out, " %s,",curr_w->wirename);
		}
	  if(new_line==10)
	  {
		fprintf(fp_out, "\n");
		new_line=0;
	  }
    }
    fprintf(fp_out, ";\n");
    fprintf(fp_out, "\n");

	cout<<" Output wire \n";

 

    int counter=1;
//Output the recognized and unrecognized CMOS gates
	Branch *my_branch;
	for(int jj=0; jj<(num_of_gates); jj++)
    {
	 if(hash_gate_list[jj]!=NULL)
	 {
	  if(hash_gate_list[jj]->cmos_gate)
	  {
		my_branch = hash_gate_list[jj]->n_branch;
//	    MosList *curmos=my_branch->mos_list;
	    fprintf(fp_out, "    %s  U%d ( ", my_branch->headername, counter);
	    counter+=1;
	    for(NameList *cur_name=my_branch->name_list; cur_name!=NULL; cur_name=cur_name->next)
	    {
		fprintf(fp_out, ".%c(%s), ", cur_name->virtual_name, cur_name->gatename);
	    }
	    fprintf(fp_out, ".out(%s) );\n", my_branch->pn_node);
/*
	    if(my_branch->greatthan4==1)
	    {
		fprintf(fp_header, "module %s(", my_branch->headername);
		for(NameList *cur_n=my_branch->name_list; cur_n!=NULL; cur_n=cur_n->next)
		    fprintf(fp_header, "%c, ", cur_n->virtual_name);
		fprintf(fp_header, "out);\n");
		fprintf(fp_header, "input ");
		for(NameList *cur_p=my_branch->name_list; cur_p!=NULL; cur_p=cur_p->next)
                    fprintf(fp_header, "%c, ", cur_p->virtual_name);
		fprintf(fp_header, ";\n");
		fprintf(fp_header, "output out;\n");
		fprintf(fp_header, "assign out = ");
		MosList *mos = my_branch->mos_list;
		fprintf(fp_header, "%s;\n", mos->g);
		fprintf(fp_header, "endmodule\n");
	    }
*/
	  }
	 }
    }

	cout<<" Output cmos_gates \n";

    fprintf(fp_out, "\n"); 
    fprintf(fp_out, "\n"); 

// Output the subcircuit list
    fprintf(fp_out, " /***The following are subcircuit : ***/\n");

	for(ResultList *curr_r=ResultHead; curr_r!=NULL; curr_r=curr_r->next)
	{
	    fprintf(fp_out, "    %s  X%d ( ", curr_r->libname, counter);
		counter += 1;
	    for(EqualNameList *cur_name=curr_r->input_list; cur_name!=NULL; cur_name=cur_name->next)
	    {
			fprintf(fp_out, ".%s(%s), ", cur_name->macromodel_name, cur_name->circuit_name);
	    }

    	for(EqualNameList *cur_2=curr_r->output_list; cur_2!=NULL; cur_2=cur_2->next)
    	{
			if(cur_2->next==NULL)
	    		fprintf(fp_out, ".%s(%s) );", cur_2->macromodel_name, cur_2->circuit_name);
			else
	    		fprintf(fp_out, ".%s(%s), ", cur_2->macromodel_name, cur_2->circuit_name);
    	}
    	fprintf(fp_out, "\n");

	}
    fprintf(fp_out, "\n"); 
    fprintf(fp_out, "\n"); 

	cout<<" Output subcircuit \n";

// Output the special_gate list

    fprintf(fp_out, " /***The following are Pseudo_Gate : ***/\n");

	for(int pp=0; pp<(num_of_gates); pp++)
    {
	 if(hash_gate_list[pp]!=NULL)
	 {
	  if(!hash_gate_list[pp]->cmos_gate)
	  {
		my_branch = hash_gate_list[pp]->n_branch;
//			fprintf(fp_out," p_log: %s, n_log: %s", (hash_gate_list[pp]->p_branch)->headername, (hash_gate_list[pp]->n_branch)->headername);
//	    MosList *curmos=my_branch->mos_list;
	    fprintf(fp_out, "    %s  U%d ( ", my_branch->headername, counter);
	    counter+=1;
	    for(NameList *cur_name=my_branch->name_list; cur_name!=NULL; cur_name=cur_name->next)
	    {
		fprintf(fp_out, ".%c(%s), ", cur_name->virtual_name, cur_name->gatename);
	    }
	    fprintf(fp_out, ".out(%s) );\n", my_branch->pn_node);
	  }
	 }
	}

    fprintf(fp_out, "\n"); 
    fprintf(fp_out, "\n"); 


//Output the pass transistor
//    fprintf(fp_out, " /***The following are transmission gates: ***/\n");
/*
	for(int xx=0; xx<(num_of_trans); xx++)
    {
	  if(hash_trans_list[xx]!=NULL)
	  {
	    fprintf(fp_out,"    TRANS U%d ( .in(%s), .out(%s) );\n", counter,hash_trans_list[xx]->input, hash_trans_list[xx]->output);
	    counter+=1;
	  }
    }
    fprintf(fp_out, "\n");
    fprintf(fp_out, "\n");
*/

//Output the non-mosfet devices

//    fprintf(fp_out, " /***The following are non-mosfet devices:***/\n");

    Element *current=ElementHead;
    PinList *curr_pin;
    while(current != NULL)
    {
        fprintf(fp_out, " %s ", current->name);

        curr_pin = current->nodelist;
        while(curr_pin != NULL)
        {
            fprintf(fp_out,"%s ", curr_pin->NodeName);
            curr_pin = curr_pin->next;
        }

        fprintf(fp_out, "%e ", current->value);

        if(current->Model != NULL)
            fprintf(fp_out, "%s ", current->Model);

		fprintf(fp_out,"\n");

        current = current->next;
    }
    fprintf(fp_out, "\n");
    fprintf(fp_out, "\n");

    fprintf(fp_out, "endmodule\n");

    time_end = UnixTime();
    time = time_end - time_begin;
    printf("******Generate gate time*********\n");
    printf("Gate gen CPU time = %f\n", gate_finish-time_begin);
    printf("******The pattern-matching time*********\n");
    printf("patt matching CPU time = %f\n", time_end-patt_begin);
    printf("******The CPU time*********\n");
    printf("CPU time = %f\n", time);
	printf("\n\n");


      times(&buffer);
      totaluser = buffer.tms_utime;
      totalsys = buffer.tms_stime;
	printf("/********** Timing Analysis ****************/\n");
    printf("Phase I Time:\n\tUser %f ", (miduser - prevuser) / 60);
    printf("\n\tSystem %f", (midsys - prevsys) / 60);
	printf("\n");
    printf("Phase II Time:\n\tUser %f ", (totaluser - miduser) / 60);
    printf("\n\tSystem %f", (totalsys - midsys) / 60);
	printf("\n");
    printf("Total Time:\n\tUser %f ", (totaluser - prevuser) / 60);
    printf("\n\tSystem %f", (totalsys - prevsys) / 60);
	printf("\n");

	printf("Parser Time: \n");
	printf("\tUser %f \n", (parseruser-prevuser)/60);
	printf("\tSystem %f \n", (parsersys-prevsys)/60);
	printf("TellNP Time: \n");
	printf("\tUser %f \n", (tellnpuser-parseruser)/60);
	printf("\tSystem %f \n", (tellnpsys-parsersys)/60);
	printf("Node Time: \n");
	printf("\tUser %f \n", (nodeuser-tellnpuser)/60);
	printf("\tSystem %f \n", (nodesys-tellnpsys)/60);
	printf("Generate Branch Time: \n");
	printf("\tUser %f \n", (branchuser-nodeuser)/60);
	printf("\tSystem %f \n", (branchsys-nodesys)/60);
	printf("Group Branch Time: \n");
	printf("\tUser %f \n", (gbranchuser-branchuser)/60);
	printf("\tSystem %f \n", (gbranchsys-branchsys)/60);
	printf("Pre Branch Recog Time: \n");
	printf("\tUser %f \n", (prerecoguser-gbranchuser)/60);
	printf("\tSystem %f \n", (prerecogsys-gbranchsys)/60);
	printf("NP Branch Recog Time: \n");
	printf("\tUser %f \n", (nprecoguser-prerecoguser)/60);
	printf("\tSystem %f \n", (nprecogsys-prerecogsys)/60);
	printf("Couple Time: \n");
	printf("\tUser %f \n", (coupleuser-nprecoguser)/60);
	printf("\tSystem %f \n", (couplesys-nprecogsys)/60);
	printf("Gate Recog Time: \n");
	printf("\tUser %f \n", (gaterecoguser-coupleuser)/60);
	printf("\tSystem %f \n", (gaterecogsys-couplesys)/60);
	printf("Pre Gateconnection Time: \n");
	printf("\tUser %f \n", (preconnectuser-gaterecoguser)/60);
	printf("\tSystem %f \n", (preconnectsys-gaterecogsys)/60);
	printf("Gate Connection Time: \n");
	printf("\tUser %f \n", (gateconnectuser-preconnectuser)/60);
	printf("\tSystem %f \n", (gateconnectsys-preconnectsys)/60);
	printf("Left Time: \n");
	printf("\tUser %f \n", (miduser-gateconnectuser)/60);
	printf("\tSystem %f \n", (midsys-gateconnectsys)/60);



    return 0;
}
	
	
double UnixTime()  /* C function returning an integer value */
{
  struct tms TimeReading;           /* tms declared in sys/times.h   */

  times(&TimeReading);              /* times declared in sys/times.h */
  return (TimeReading.tms_utime/60.0); /* returning the user time field */
                                    /* in unit of a second          */
}


MacroModelList *return_macromodel()
{
	int cou=0;
	MacroModelList *re = NULL;
	for(MacroModelList *curr=MacroModelHead; curr!=NULL; curr=curr->next)
	{
		if((curr->num_of_mos > cou) && !curr->done)
		{
			cou = curr->num_of_mos;
			re = curr;
		}
	}
	re->done = true;
	return re;
}


bool macromodel_head_done()
{
	for(MacroModelList *curr=MacroModelHead; curr!=NULL; curr=curr->next)
	{
		if(!curr->done)
			return false;
	}
	return true;
}
			
	    
