#include "pre.h"
#include "branch.h"
#include "group.h"
#include <malloc.h>
#include <stdio.h>
#include <string.h>

#define NLAlloc(type,num) (type *)calloc(num,sizeof(type))

extern MosList *MosHead;
extern NodeList *NodeHead;

Group::Group()
{
	p_branch_list = NULL;
	n_branch_list = NULL;
	pass_transistor_list = NULL;
	left_transistor_list = NULL;
}

Group::~Group()
{
	if(p_branch_list) delete p_branch_list;
	if(n_branch_list) delete n_branch_list;
	if(pass_transistor_list) delete pass_transistor_list;
	if(left_transistor_list) delete left_transistor_list;
}

char *
Group::get_vdd(MosList *moshead)
{
   char *vdd;
   MosList *curr = NULL;
   for(curr=moshead; curr!=NULL; curr=curr->next)
   {
        if(curr->nptype)
        {
          vdd = NLAlloc(char,strlen(curr->b));
          strcpy(vdd, curr->b);
          break;
        }
   }
   return vdd;
}

char *
Group::get_gnd(MosList *moshead)
{
   char *gnd;
   MosList *curr = NULL;
   for(curr=moshead; curr!=NULL; curr=curr->next)
   {
        if(!curr->nptype)
        {
          gnd = NLAlloc(char,strlen(curr->b));
          strcpy(gnd, curr->b);
          break;
        }
   }
   return gnd;
}

int
Group::query_n_branch(char *name)
{
   int exist_flag = 0;
   if(n_branch_list==NULL)
	return 0;
   for(Branch *cur_branch=n_branch_list; cur_branch!=NULL; cur_branch=cur_branch->next)
   {
	for(MosList *cur_mos=cur_branch->mos_list; cur_mos!=NULL; cur_mos=cur_mos->next)
	{
	   if(!strcmp(cur_mos->mosname, name))
	   {
		exist_flag = 1;
		return 1;
	   }
	}
   }
   
   if(!exist_flag)
	return 0;
}

int
Group::query_p_branch(char *name)
{
   int exist_flag = 0;
   if(p_branch_list==NULL)
	return 0;
   for(Branch *cur_branch=p_branch_list; cur_branch!=NULL; cur_branch=cur_branch->next)
   {
        for(MosList *cur_mos=cur_branch->mos_list; cur_mos!=NULL; cur_mos=cur_mos->next)
        {
           if(!strcmp(cur_mos->mosname, name))
	   {
              exist_flag = 1;
	      return 1;
	   }
        }
   }

   if(!exist_flag)
        return 0;
}

Branch *
Group::insert_branch(Branch *head, Branch *new_branch)
{
   Branch *current;
   if(head==NULL)
   {
	head = new_branch;
	head->next = NULL;
   }
   else
   {
	current = new_branch;
	current->next = head->next;
	head->next = current;
   }
   return head;
}

void
Group::generate_branch_list()
{
   char *gnd;
   gnd = get_gnd(MosHead);
   char *vdd;
   vdd = get_vdd(MosHead);

   NodeList *gnd_node, *vdd_node;

   for(NodeList *curr_node=NodeHead; curr_node!=NULL; curr_node=curr_node->next)
   {
	if(!strcmp(curr_node->nodename, gnd))
	   gnd_node = curr_node;
	if(!strcmp(curr_node->nodename, vdd))
	   vdd_node = curr_node;
   }


   for(DeviceList *cur_device=gnd_node->devicelist; cur_device!=NULL; cur_device=cur_device->next)
   {
//	printf("%s \n", cur_device->devicename);
	MosList *first_n_mos;
	if(!query_n_branch(cur_device->devicename))	
	{
	   Branch *new_branch;
	   new_branch = new Branch(0);	

	   for(MosList *cur_m=MosHead; cur_m!=NULL; cur_m=cur_m->next)
	   {
		if(!strcmp(cur_device->devicename, cur_m->mosname))
		{
		   first_n_mos = cur_m;
//		   printf("A new branch from this mos: %s \n", cur_device->devicename);
		   break;
		}
	   }

	   new_branch->generate_branch(first_n_mos);
/*
	   for(MosList *cur_b=new_branch->mos_list; cur_b!=NULL; cur_b=cur_b->next)
		printf("%s  ", cur_b->mosname);
	   printf("\n");	
*/
	   n_branch_list = insert_branch(n_branch_list, new_branch);
	}
   }

   for(DeviceList *cur_device=vdd_node->devicelist; cur_device!=NULL; cur_device=cur_device->next)
   {
        MosList *first_p_mos;
        if(!query_p_branch(cur_device->devicename))
        {
           Branch *new_branch;
           new_branch = new Branch(1);

           for(MosList *cur_m=MosHead; cur_m!=NULL; cur_m=cur_m->next)
           {
                if(!strcmp(cur_device->devicename, cur_m->mosname))
                {
                   first_p_mos = cur_m;
//	           printf("A new branch from this mos: %s \n", cur_device->devicename);
                   break;
                }
           }

           new_branch->generate_branch(first_p_mos);
/*
           for(MosList *cur_b=new_branch->mos_list; cur_b!=NULL; cur_b=cur_b->next)
                printf("%s  ", cur_b->mosname);
           printf("\n");
*/
           p_branch_list = insert_branch(p_branch_list, new_branch);
	}
   }
}

void
Group::group_branch_list()
{
   if(p_branch_list->next != NULL)
   {
     for(Branch *cur_branch=p_branch_list; cur_branch!=NULL; cur_branch=cur_branch->next)
     {
//		MosList *cu=cur_branch->mos_list;
//	        printf("Deal with: %s \n",cu->mosname); 
	Branch *curr, *prev, *temp;
	prev=cur_branch;
	curr=cur_branch->next;	
	while(curr!=NULL)
	{
	   if(!strcmp(curr->pn_node, cur_branch->pn_node))
	   {
		for(MosList *cur_mos=curr->mos_list; cur_mos!=NULL; cur_mos=cur_mos->next)
		   cur_branch->mos_list = cur_branch->insert_mos(cur_branch->mos_list,cur_mos);
		for(BranchNodeList *cur_node=curr->node_list; cur_node!=NULL; cur_node=cur_node->next)
		   cur_branch->node_list = cur_branch->insert_node(cur_branch->node_list,cur_node->nodename);
		prev->next = curr->next;
		curr = prev->next;
/*
		temp = prev->next;
		prev = temp;
		curr = temp->next;
*/
	   }
	   else
	   {
	   prev=curr;
	   curr=curr->next;
	   }
	}
/*
	for(Branch *cur_br=p_branch_list; cur_br!=NULL; cur_br=cur_br->next)
	{
		MosList *cu1=cur_br->mos_list;
	        printf("%s ",cu1->mosname); 
	}
	printf("\n");
*/
      }
   }



   if(n_branch_list->next != NULL)
   {
     for(Branch *cur_branch=n_branch_list; cur_branch!=NULL; cur_branch=cur_branch->next)
     {
        Branch *curr, *prev, *temp;
        prev=cur_branch;
        curr=cur_branch->next;
        while(curr!=NULL)
        {
           if(!strcmp(curr->pn_node, cur_branch->pn_node))
           {
                for(MosList *cur_mos=curr->mos_list; cur_mos!=NULL; cur_mos=cur_mos->next)
                   cur_branch->mos_list = cur_branch->insert_mos(cur_branch->mos_list,cur_mos);
		for(BranchNodeList *cur_node=curr->node_list; cur_node!=NULL; cur_node=cur_node->next)
                   cur_branch->node_list = cur_branch->insert_node(cur_branch->node_list,cur_node->nodename);
                prev->next = curr->next;
		curr = prev->next;
/*
                temp = prev->next;
                prev = temp;
                curr = temp->next;
*/
           }
           else
           {
           prev=curr;
           curr=curr->next;
           }
        }
/*
        for(Branch *cur_br=n_branch_list; cur_br!=NULL; cur_br=cur_br->next)
        {
                MosList *cu1=cur_br->mos_list;
                printf("%s ",cu1->mosname);
        }
        printf("\n");
*/
      } 
   }    
}

MosList *
Group::insert_mos(MosList *head, MosList *mos)
{
   MosList *current;
   if(head==NULL)
   {
        head = NLAlloc(MosList,1);
        head->mosname = NLAlloc(char,strlen(mos->mosname));
        strcpy(head->mosname,mos->mosname);
        head->d = NLAlloc(char,strlen(mos->d));
        strcpy(head->d,mos->d);
        head->g = NLAlloc(char,strlen(mos->g));
        strcpy(head->g,mos->g);
        head->s = NLAlloc(char,strlen(mos->s));
        strcpy(head->s,mos->s);
        head->nptype = mos->nptype;
        head->b = NLAlloc(char,strlen(mos->b));
        strcpy(head->b,mos->b);
        head->modelname = NLAlloc(char,strlen(mos->modelname));
        strcpy(head->modelname,mos->modelname);
        head->lvalue = mos->lvalue;
        head->wvalue = mos->wvalue;
        head->next = NULL;
   }
   else
   {
        current = NLAlloc(MosList,1);
        current->mosname = NLAlloc(char, strlen(mos->mosname));
        strcpy(current->mosname,mos->mosname);
        current->d = NLAlloc(char, strlen(mos->d));
        strcpy(current->d,mos->d);
        current->g = NLAlloc(char, strlen(mos->g));
        strcpy(current->g,mos->g);
        current->s = NLAlloc(char, strlen(mos->s));
        strcpy(current->s,mos->s);
        current->nptype = mos->nptype;
        current->b = NLAlloc(char, strlen(mos->b));
        strcpy(current->b,mos->b);
        current->modelname = NLAlloc(char, strlen(mos->modelname));
        strcpy(current->modelname,mos->modelname);
        current->lvalue = mos->lvalue;
        current->wvalue = mos->wvalue;
        current->next = head->next;
        head->next = current;
   }
   return head;
}


void
Group::exchange_d_s(MosList *mos)
{
    char *temp=NLAlloc(char, strlen(mos->d));
    strcpy(temp, mos->d);
    mos->d = mos->s;
    mos->s = temp;
}

int
Group::query_is_pnnode(char *name)
{
    for(NodeList *cur_n=NodeHead; cur_n!=NULL; cur_n=cur_n->next)
    {
	if(!strcmp(cur_n->nodename,name))
	{
	    return cur_n->pn_node;
	}
    }
}

int
Group::query_is_gnode(char *name)
{
    int p=0;
    for(NodeList *cur_n=NodeHead; cur_n!=NULL; cur_n=cur_n->next)
    {
        if(!strcmp(cur_n->nodename,name))
        {
            for(DeviceList *cur_d=cur_n->devicelist; cur_d!=NULL; cur_d=cur_d->next)
            {
                if(cur_d->pintype==GP)
                {
                    p=1;
                }
            }
        }
    }
    return p;
}


void
Group::generate_pass_left_list()
{
    int is_pn_1, is_g_1;
    int is_pn_2, is_g_2;

    for(MosList *curr_mos=MosHead; curr_mos!=NULL; curr_mos=curr_mos->next)
    {
	if(!query_n_branch(curr_mos->mosname) && !query_p_branch(curr_mos->mosname))
	{
	    if((query_is_pnnode(curr_mos->d) && query_is_gnode(curr_mos->s)) || (query_is_pnnode(curr_mos->s) && query_is_gnode(curr_mos->d)))
		pass_transistor_list = insert_mos(pass_transistor_list,curr_mos);
            else
                left_transistor_list = insert_mos(left_transistor_list, curr_mos);
	}
    }	
}

/*
	    printf("%s  ", curr_mos->mosname);
	    for(NodeList *curr_node=NodeHead; curr_node!=NULL; curr_node=curr_node->next)
	    {
		if(!strcmp(curr_mos->d, curr_node->nodename))
		{
		    if(curr_node->pn_node)
		    {
			for(NodeList *curr_node2=NodeHead; curr_node2!=NULL; curr_node2=curr_node2->next)
			{
			    if(!strcmp(curr_mos->s, curr_node2->nodename))
			    {
				DeviceList *temp=curr_node2->devicelist;
				if(temp->next==NULL && temp->pintype==GP)
					pass_transistor_list = insert_mos(pass_transistor_list,curr_mos);
				else
					left_transistor_list = insert_mos(left_transistor_list, curr_mos);
				break;
			    }
			}
		    }
		    else
			left_transistor_list = insert_mos(left_transistor_list, curr_mos);
		    break;
		}

		if(!strcmp(curr_mos->s, curr_node->nodename))
                {
                    if(curr_node->pn_node)
                    {
                        for(NodeList *curr_node3=NodeHead; curr_node3!=NULL; curr_node3=curr_node3->next)
                        {
                            if(!strcmp(curr_mos->d, curr_node3->nodename))
                            {
                                DeviceList *temp=curr_node3->devicelist;
                                if(temp->next==NULL && temp->pintype==GP)
				{
					exchange_d_s(curr_mos);
                                        pass_transistor_list = insert_mos(pass_transistor_list,curr_mos);
				}
                                else
                                        left_transistor_list = insert_mos(left_transistor_list, curr_mos);
                                break;
                            }
                        }
                    }
                    else
                        left_transistor_list = insert_mos(left_transistor_list, curr_mos);
                    break;
                }

	    }
	}
	else
	    left_transistor_list = insert_mos(left_transistor_list, curr_mos);
    }
}
*/
	    
int
Group::counter(Branch *branch)
{
   int num = 0;
   for(MosList *cur_mos=branch->mos_list; cur_mos!=NULL; cur_mos=cur_mos->next)
   {
	num = num+1;
   }
   return num;
}

char *
Group::generate_function(char *g1, char *g2, int and_or_flag)
{
    const char *left = "(";
    const char *right = ")";
    const char *add = "+";
    const char *mult = "*";
    char *new_string=NLAlloc(char,strlen(left));
    strcpy(new_string,left);
    if(!and_or_flag)
    {
	 strcat(new_string, g1);
	 strcat(new_string, mult);
	 strcat(new_string, g2);
	 strcat(new_string, right);
    }
    else
    {
         strcat(new_string, g1);
         strcat(new_string, add);
         strcat(new_string, g2);
         strcat(new_string, right);
    }
    return new_string;
}
    

void
Group::group_step(Branch *p_branch, Branch *n_branch, int and_or_flag, int succeed)
{
    for(MosList *curr_mos=p_branch->mos_list; curr_mos!=NULL; curr_mos=curr_mos->next)
    {
	MosList *curr_pointer, *prev_pointer;
	prev_pointer=curr_mos;
	curr_pointer=curr_mos->next;
	while(curr_pointer!=NULL)
	{
	    if((!strcmp(curr_mos->d, curr_pointer->d) && !strcmp(curr_mos->s, curr_pointer->s)) || (!strcmp(curr_mos->d, curr_pointer->s) && !strcmp(curr_mos->s, curr_pointer->d)))
	    {
		MosList *nmos1;
		int flag1=0;
		for(MosList *curr_mos2=n_branch->mos_list; curr_mos2!=NULL; curr_mos2=curr_mos2->next)
		{
		    if(!strcmp(curr_mos->g, curr_mos2->g))
		    {
			nmos1=curr_mos2;
			flag1=1;
			break;
		    }
		}
		if(!flag1)
		{
		    succeed = 0;
		}

		int flag2 =0;
		MosList *temp;
		MosList *curr_nmos, *prev_nmos;
		prev_nmos = n_branch->mos_list;
		curr_nmos = n_branch->mos_list;
		while(curr_nmos!=NULL)
		{
		    if(!strcmp(curr_nmos->g, curr_pointer->g))	
		    {
			flag2 = 1;
			if((!strcmp(nmos1->d, curr_nmos->d) && strcmp(nmos1->s, curr_nmos->s)) || (!strcmp(nmos1->s, curr_nmos->s) && strcmp(nmos1->d, curr_nmos->d)) || (!strcmp(nmos1->d, curr_nmos->s) && strcmp(nmos1->s, curr_nmos->d)) || (!strcmp(nmos1->s, curr_nmos->d) && strcmp(nmos1->d, curr_nmos->s)))	
			{

			   curr_mos->g = generate_function(curr_mos->g, curr_pointer->g, and_or_flag);
			   prev_pointer->next = curr_pointer->next;
			   temp = prev_pointer->next;
			   prev_pointer = temp;
			   curr_pointer = temp->next;

			   nmos1->g = generate_function(nmos1->g,curr_nmos->g,and_or_flag);
			   if(!strcmp(nmos1->d, curr_nmos->d) && strcmp(nmos1->s, curr_nmos->s))
				strcpy(nmos1->d, curr_nmos->s);
			   if(!strcmp(nmos1->s, curr_nmos->s) && strcmp(nmos1->d, curr_nmos->d))
				strcpy(nmos1->s, curr_nmos->d);
			   if(!strcmp(nmos1->d, curr_nmos->s) && strcmp(nmos1->s, curr_nmos->d)) 
				strcpy(nmos1->d, curr_nmos->d);
			   if(!strcmp(nmos1->s, curr_nmos->d) && strcmp(nmos1->d, curr_nmos->s))
				strcpy(nmos1->s, curr_nmos->s);
			   prev_nmos->next = curr_nmos->next;
			   delete curr_nmos;
			   break;
			}
			else
			{
			   succeed = 0;
			}
		    }
		    else
		    {
			prev_nmos = curr_nmos;
			curr_nmos = curr_nmos->next;
		    }
		}	    
		if(!flag2)
		{
		    succeed = 0;
		}
	    }
	    else
	    {
		prev_pointer = curr_pointer;
		curr_pointer = curr_pointer->next;
	    }
	}
    }
}

void
Group::branch_recognition()
{
    for(Branch *curr_p=p_branch_list; curr_p!=NULL; curr_p=curr_p->next)
    {
	for(Branch *curr_n=n_branch_list; curr_n!=NULL; curr_n=curr_n->next)
	{
	    if(!strcmp(curr_p->pn_node, curr_n->pn_node))
	    {
		if(counter(curr_p)!=counter(curr_n))
		{
 		    curr_p->success = 0;
		    curr_n->success = 0;
   		}
		else
   		{
		    if(counter(curr_p) == 1)
		    {
	   		curr_p->success = 1;
	   		curr_n->success = 1;
		    }
		    else
		    {
			MosList *list=curr_p->mos_list;
			while(list->next != NULL)
			{
			    int succeed1 = 1;
			    group_step(curr_p,curr_n,0,succeed1); 
			    if(!succeed1)
			    {
			        curr_p->success = 0;
			        curr_n->success = 0;
			        break;
			    }
			   
			    int succeed2 = 1;
			    group_step(curr_n, curr_p, 1, succeed2);
			    if(!succeed2)
			    {
				curr_p->success = 0;
                                curr_n->success = 0;
                                break;
			    }
			}
			if(list->next == NULL)
			{
			    curr_p->success = 1;
                            curr_n->success = 1;
			}
		    }
		}
	    }
	}
    }
}
				

